;
; Action 53 title screen and game loader
; Copyright 2011-2014 Damian Yerrick
;
; Copying and distribution of this file, with or without
; modification, are permitted in any medium without royalty provided
; the copyright notice and this notice are preserved in all source
; code copies.  This file is offered as-is, without any warranty.
;

.include "nes.inc"
.include "global.inc"
.include "pently.inc"

A53_MAPPER = 1

.import cart_menu, get_titledir_a
.import clearLineImg, copyLineImg, vwfPuts
.import interbank_fetch, interbank_fetch_buf
.import quadpcm_test
.import coredump
.importzp lineImgBufLen
.exportzp psg_sfx_state

; ld65 manual says LOWCODE is for stuff guaranteed not to be
; banked out, good for irq handlers, but it's also good for
; anything that must access data from multiple banks
.import __LOWCODE_RUN__, __LOWCODE_LOAD__
.import __LOWCODE_SIZE__

.segment "ZEROPAGE"
nmis:          .res 1
oam_used:      .res 1

; Used by pads.s
cur_keys:  .res 2
new_keys:  .res 2
cur_trigger: .res 1

; Used by music engine
psg_sfx_state: .res 32
tvSystem:   .res 1

.segment "BSS"
linebuf: .res 40
ciBlocksLeft = linebuf

.if 0
.segment "INESHDR"
  .byt "NES",$1A  ; magic signature
  .byt 2          ; PRG ROM size in 16384 byte units
  .byt 0          ; CHR ROM size in 8192 byte units
  .byt $00        ; mirroring type and mapper number lower nibble
  .byt $00        ; mapper number upper nibble
.endif

.segment "BFF0"
.proc patch_bff0
  sei
ldxinstr:
  ldx #$FF
  nop
  stx ldxinstr+1
  jmp ($FFFC)
.endproc

.segment "FFF0"
.proc patch_fff0
  sei
ldxinstr:
  ldx #$FF
  nop
  stx ldxinstr+1
  jmp ($FFFC)
  .addr nmi, reset, irq
.endproc

; Add this only when copying hex into the patcher
.if 1
.segment "CODE"
.proc patch21
  sei
  ldx #copied_end - copied - 1
:
  lda copied,x
  sta $00,x
  dex
  bpl :-
  jmp $0000
copied:
  stx copied+5
  jmp ($FFFC)
copied_end:
.endproc
.endif

.segment "CODE"
;;
; This NMI handler is good enough for a simple "has NMI occurred?"
; vblank-detect loop, like in Lawn Mower or Thwaite.
.proc nmi
  inc nmis
  rti
.endproc

; BRK $00 to add a breakpoint that survives rebuild
.proc irq
  rti
.endproc

; 
.proc reset
  ; The very first thing to do when powering on is to put all sources
  ; of interrupts into a known state.
  sei             ; Disable interrupts
  ldx #$00
  stx PPUCTRL     ; Disable NMI and set VRAM increment to 32
  stx PPUMASK     ; Disable rendering
  stx $4010       ; Disable DMC IRQ
  dex             ; Leave for later TXS
  bit PPUSTATUS   ; Acknowledge stray vblank NMI across reset
  bit SNDCHN      ; Acknowledge DMC IRQ
  lda #$40
  sta P2          ; Disable APU Frame IRQ
  lda #$0F
  sta SNDCHN      ; Disable DMC playback, initialize other channels

vwait1:
  bit PPUSTATUS   ; It takes one full frame for the PPU to become
  bpl vwait1      ; stable.  Wait for the first frame's vblank.

coredump_at_boot_readpad:
  ldy #$01
  lda #$00
  sty $4016
  sta $4016
  @readpad_loop:
    lda $4017
    ora $4016
    and #%00000011  ; ignore D2-D7
    cmp #1          ; CLC if A=0, SEC if A>=1
    tya
    rol
    tay
  bcc @readpad_loop
  ; A = Y = pad2
  cpy #%11000000
  bne @not_coredump
    jmp coredump
  @not_coredump:

  ; Set stack pointer now that we know we're not running coredump
  txs

  ; Turn off decimal mode for post-patent famiclones
  cld

  ; Copy the CHR decompression code to RAM.
  .assert __LOWCODE_SIZE__ < 256, error, "LOWCODE too big"
  ldx #0
copy_LOWCODE:
  lda __LOWCODE_LOAD__,x
  sta __LOWCODE_RUN__,x
  inx
  cpx #<__LOWCODE_SIZE__
  bne copy_LOWCODE

  jsr pently_init

  .if ::A53_MAPPER
    jsr init_mapper28
  .endif

vwait2:
  bit PPUSTATUS  ; After the second vblank, we know the PPU has
  bpl vwait2     ; fully stabilized.  After this use only NMI.

  lda #VBLANK_NMI
  sta PPUCTRL
  jsr getTVSystem
  sta tvSystem
  
  lda TITLESCREEN
  cmp #$FF
  bne has_games
    jmp no_games_error
  has_games:

  lda #0
  jsr pently_start_music
  jsr title_screen
  jsr ppu_wait_vblank
  jsr cart_menu
  ldx #0
  stx SNDCHN
  ; game id is in A
  pha
  jsr get_titledir_a
  jsr load_titledir_chr_rom
  ; Snail Maze Game doesn't properly clear OAM.
  ldx #0
  jsr ppu_clear_oam
  pla
  jsr get_titledir_a
  jmp jump_to_titledir
.endproc

line1dst = $0D08
.segment "CODE"
.proc title_screen
  bit PPUSTATUS
  lda #VBLANK_NMI
  sta PPUCTRL
  
  lda TITLESCREEN+1
  sta ciSrc+1
  lda TITLESCREEN+0
  sta ciSrc
  
  ; Unpack tiles
  ldy #0
  sty PPUMASK
  sty PPUADDR
  sty PPUADDR
  lda (ciSrc),y
  tax
  inc ciSrc
  bne :+
    inc ciSrc+1
  :
  jsr unpb53_block

  ; 48 blank tiles for copyright notice
  lda #$0D
  sta PPUADDR
  lda #$00
  sta PPUADDR
  tax
lastlineclr:
  sta PPUDATA
  inx
  bne lastlineclr
  dey
  bne lastlineclr

  ; Unpack nametable
  lda #$20
  sta PPUADDR
  lda #$00
  sta PPUADDR
  ldx #1024/16
  jsr unpb53_block

  ; fill in the palette
  ldx #$3F
  stx PPUADDR
  sta PPUADDR
  lda nmis
  :
    cmp nmis
    beq :-
palloop:
  lda (ciSrc),y
  sta PPUDATA
  iny
  cpy #16
  bcc palloop

  ; Draw the serial number and copyright notices
  lda #0
  sta draw_progress
copr_notices_loop:
  clc
  adc #$0D
  pha  ; stack: high byte of copy dst, ...
  asl a
  asl a
  asl a
  asl a
  tay
  lda draw_progress
  asl a
  asl a
  adc draw_progress
  pha  ; stack: draw_progress * 5, high byte of copy dst, ...
  tax
  lda title_msgs+1,x
  sta PPUADDR
  lda title_msgs+0,x
  sta PPUADDR
  ldx #lineImgBufLen/8
  lda #VBLANK_NMI
  sta PPUCTRL
mknt2:
  sty PPUDATA
  iny
  dex
  bne mknt2
  jsr clearLineImg
  pla
  tax
  
  ldy title_msgs+2,x
  lda title_msgs+3,x
  pha  ; stack: high byte of text src, high byte of copy dst, ...
  lda title_msgs+4,x
  tax
  pla
  jsr vwfPuts
  pla
  ldy #$08
  jsr copyLineImg
  
  inc draw_progress
  lda draw_progress
  cmp #3
  bcc copr_notices_loop

title_wait_A:
  lda #VBLANK_NMI|BG_0000
  sta PPUCTRL
  jsr ppu_wait_vblank
  lda #$00
  sta PPUSCROLL
  sta PPUSCROLL
  lda #BG_ON
  sta PPUMASK
  jsr read_pads
  ldx #0
  jsr read_mouse_with_backward_buttons
  jsr pently_update
  lda pently_music_playing
  beq selnow
  jsr read_zapper_trigger
  ora new_keys
  and #KEY_START|KEY_A
  beq title_wait_A
  rts
selnow:
  lda #0
  sta PPUMASK
  jmp quadpcm_test
.endproc

.proc no_games_error
  lda #0
  tay
  tax
  jsr ppu_clear_nt
  lda #0
  tay
  ldx #$20
  jsr ppu_clear_nt
  lda #$21
  sta PPUADDR
  lda #$09
  sta PPUADDR
  ldx #$01
  :
    stx PPUDATA
    inx
    cpx #16
    bcc :-
  lda #$3F
  sta PPUADDR
  lda #$00
  sta PPUADDR
  lda #$16
  sta PPUDATA
  lda #$0F
  sta PPUDATA
  jsr clearLineImg
  lda #>no_games_msg
  ldy #<no_games_msg
  ldx #8
  jsr vwfPuts
  lda #$00
  tay
  jsr copyLineImg
  ldx #0
  ldy #0
  lda #VBLANK_NMI
  clc
  jmp ppu_screen_on
.endproc

.segment "RODATA"
title_msgs:
  .addr $22E8, serial_msg
  .byte 48
  .addr $2326, copr_msg_1
  .byte 4
  .addr $2334, copr_msg_2
  .byte 0
serial_msg: .byte "Page 8",0
copr_msg_1: .byte "2017 Damian Yerrick and",0
copr_msg_2: .byte "Contributors",0
no_games_msg: .byte "No games added.",0

.importzp ciSrc, ciDst, ciBufStart, ciBufEnd
.importzp ciSrc0, ciSrc1
.import draw_progress
.import unpb53_some, PB53_outbuf, interbank_fetch, interbank_fetch_buf

.segment "CODE"

;;
; decompress X*16 bytes starting at AAYY to PPUDATA
.proc unpb53_block_ay
  sty ciSrc
  sta ciSrc+1
.endproc
.proc unpb53_block
  stx draw_progress
  lda #16
  sta ciBufEnd
  lda #0
  sta ciBufStart
loop:
  jsr unpb53_some
  ldx #0
copyloop:
  lda PB53_outbuf,x
  sta PPUDATA
  inx
  cpx #16
  bcc copyloop
  dec draw_progress
  bne loop
  rts
.endproc

; Temporary measure during debugging to keep Sinking Feeling-only ROM
; from reading off the end.  Once this is working, I'll try adding
; the CHR ROM size to the title directory.
CNROM_MAX_SIZE = 2

;;
; Loads the CHR bank ID associated with this title.
; @param $0000 pointer to entry in title directory
; @param $FF02 pointer to start of CHR directory, where each entry is
; 5 bytes: PRG bank, address low, high, midpoint offset low, high
.proc load_titledir_chr_rom
titleptr = $00
chrdir_entry_zp = $02  ; matches interbank_fetch bank ptr
chrdir_entry = interbank_fetch_buf+72
cur_chr_bank = interbank_fetch_buf+74
num_chr_banks = interbank_fetch_buf+75

  lda #1
  sta num_chr_banks
  lsr a
  sta cur_chr_bank
  sta PPUMASK
  sta PPUCTRL
  ldy #TITLE_CHR_BANK
  lda (titleptr),y
  bpl is_chr_rom  ; CHR bank >= $80 means absent
    rts
  is_chr_rom:

  ; A = starting CHR bank number, 0 to 127.
  ; Calculate offset into chrdir as 5 * A + CHRDIR_START
  .if ::TITLE_CHR_BANK <> 1
    ldy #1
  .endif
  dey
  sty $4444
  sty chrdir_entry_zp+1
  sta chrdir_entry_zp
  asl a
  asl a
  rol chrdir_entry_zp+1
  adc chrdir_entry_zp
  bcc :+
    inc chrdir_entry_zp+1
    clc
  :
  adc CHRDIR_START
  sta chrdir_entry
  lda CHRDIR_START+1
  adc chrdir_entry_zp+1
  sta chrdir_entry+1

  ldy #TITLE_MAPPER_CFG
  lda (titleptr),y  ; bit 7: 0 for CNROM, 1 for not
  and #$80
  bmi not_cnrom
  ldy #CNROM_MAX_SIZE
  sty num_chr_banks
nextbank:
  lda #$00
  sta $5000
  lda cur_chr_bank
  sta $8000
  inc cur_chr_bank
  lda #$81
  sta $5000
not_cnrom:
  lda chrdir_entry
  sta chrdir_entry_zp
  lda chrdir_entry+1
  sta chrdir_entry_zp+1
  ldy #0
  sty draw_progress
  iny
  lda (chrdir_entry_zp),y
  iny
  sta ciSrc0
  lda (chrdir_entry_zp),y
  iny
  sta ciSrc0+1
  clc
  lda (chrdir_entry_zp),y
  iny
  adc ciSrc0
  sta ciSrc1
  lda (chrdir_entry_zp),y
  adc ciSrc0+1
  sta ciSrc1+1

  ; At this point:
  ; 2: pointer to bank number
  ; ciSrc0: Address of first 4 KiB
  ; ciSrc1: Address of second 4 KiB
  ; drawProgress: number of blocks already drawn
loop:
  lda ciSrc0
  sta 0
  lda ciSrc0+1
  sta 1
  lda #0
  jsr do4
  adc ciSrc0
  sta ciSrc0
  bcc :+
  inc ciSrc0+1
:

  lda ciSrc1
  sta 0
  lda ciSrc1+1
  sta 1
  lda #64
  jsr do4
  adc ciSrc1
  sta ciSrc1
  bcc :+
  inc ciSrc1+1
:
  lda #0
  sta 0
  lda draw_progress
  lsr a
  ror 0
  lsr a
  ror 0
  sta PPUADDR
  ora #$10
  sta 1
  lda 0
  sta PPUADDR
  ldx #0
copyloop0:
  lda PB53_outbuf,x
  sta PPUDATA
  inx
  cpx #64
  bcc copyloop0
  lda 1
  sta PPUADDR
  lda 0
  sta PPUADDR
copyloop1:
  lda PB53_outbuf,x
  sta PPUDATA
  inx
  cpx #128
  bcc copyloop1
  inc draw_progress
  lda draw_progress
  cmp #64
  bcc loop
  dec num_chr_banks
  beq no_more_chr_banks
    clc
    lda #5
    adc chrdir_entry
    sta chrdir_entry
    bcc :+
      inc chrdir_entry+1
    :
    jmp nextbank
  no_more_chr_banks:
  rts

;;
; @param A start of buffer (0 or 64)
; @return A number of compressed bytes read
do4:
  sta ciBufStart
  clc
  adc #64
  sta ciBufEnd
  lda #72  ; 4 tiles, max 18 bytes per tile
  sta 4
  lda chrdir_entry
  sta chrdir_entry_zp
  lda chrdir_entry+1
  sta chrdir_entry_zp+1
  jsr interbank_fetch
  lda #<interbank_fetch_buf
  sta ciSrc
  lda #>interbank_fetch_buf
  sta ciSrc+1
  jsr unpb53_some
  lda ciSrc
  sec
  sbc #<interbank_fetch_buf
  clc
  rts
.endproc

;;
; Reads a mouse and maps presses of L to A and R to B
; in new_keys.
; @param X port number of mouse to read (0 or 1)
; @return A zero iff mouse is present
.proc read_mouse_with_backward_buttons
  jsr read_mouse
  lda 1
  eor #$01
  and #$0F
  bne no_mouse

  lda new_mbuttons,x
  lsr a
  and #KEY_B
  ora new_keys,x
  sta new_keys,x
  lda new_mbuttons,x
  asl a
  and #KEY_A
  ora new_keys,x
  sta new_keys,x

  lda #0
no_mouse:
  rts
.endproc

.proc read_zapper_trigger
  lda $4017
  and #$10
  sta 0
  lda cur_trigger
  eor #$10
  and 0
  ldy 0
  sty cur_trigger
  rts
.endproc

.segment "LOWCODE"
.proc jump_to_titledir
titleptr = 0
resetlo = 2
resethi = 3
mapperconfig = 4

  ; Relevant entries:
  ; 0: Outer bank number
  ; 12-13: Reset vector
  ; 14: mapper 20 config
  ; (bits 5-0
  ldy #TITLE_ENTRY_POINT
  lda (titleptr),y
  sta resetlo
  iny
  lda (titleptr),y
  sta resethi
  .if ::A53_MAPPER
    iny
    lda (0),y
    sta mapperconfig
  .endif
  ldy #0
  lda (0),y
  sta (0),y

  ; Now the outer bank and reset vector are correct, and the mapper
  ; configuration has been saved in a variable.  Now interpret the
  ; mapper configuration in reg $80 format:
  ; 76543210
  ; | ||||||
  ; | ||||++- Nametable mirroring (0=AAAA, 2=ABAB, 3=AABB)
  ; | ||++--- PRG bank mode (0=32k, 2=fixed $8000, 3=fixed $C000)
  ; | ++----- Game size (0=32k, 1=64k, 2=128k, 3=256k)
  ; +-------- If set, game isn't CNROM
  ; Rules for setting final reg in $5000:
  ; If CNROM, and CHR loader supports CNROM, use $00 (CHR bank).
  ; If game is 32K, use $81 (outer PRG bank) so that simple
  ; reset code works.
  ; Otherwise, use $01 (inner PRG bank).

  .if ::A53_MAPPER
    ldx #$80
    stx $5000
    lda mapperconfig
    sta $8000
    
    ; Set the inner bank: $00 for fixed-lo or $0F otherwise
    ldy #$01
    sty $5000
    dey
    and #$0C
    cmp #$08
    beq :+
      ldy #$0F
    :
    sty $8000
    
    ; Set the reg that the program writes to.
    ; For CNROM, use $00 (CHR bank)
    ldx #$00
    lda mapperconfig
    bpl have_regnum_in_x
    
    ; For games larger than 32K, use $01 (PRG inner bank)
    inx
    and #$30
    bne have_regnum_in_x
    
    ; Otherwise use $81 (PRG outer bank)
    ldx #$81
  have_regnum_in_x:
    stx $5000
  .endif

  ; and launch.
  jmp ($0002)
.endproc

.if A53_MAPPER
.proc init_mapper28
  ldx #$80
  stx $5000
  lda #$02  ; 32k outer, bnrom style, vertical mirroring
  sta $8000
  inx
  stx $5000
  lda #$FF
  sta patch_fff0::ldxinstr+1
  rts
.endproc
.endif

; historic a53 volumes
; Emulator: http://www.fceux.com/web/download.html
; Volume 1: https://forums.nesdev.com/viewtopic.php?p=109907#p109907
; Volume 2: https://forums.nesdev.com/viewtopic.php?p=148343#p148343
; Volume 3 (work in progress): https://forums.nesdev.com/download/file.php?id=8115
