#!/usr/bin/env python3
from savtool import bisqpal as nespal

def nespal_closest(rgb):
    """Find the closest color in the 52-color NES palette to a given RGB tuple."""

    # Start with black ($0F), defined as (0, 0, 0)
    r, g, b = rgb
    bestColor = 0x0F
    bestDist = r*r*3 + g*g*6 + b*b
    for i, (nr, ng, nb) in enumerate(nespal):
        # Want to skip $xD, $xE, $xF except for $0F, which
        # we already covered
        if (i & 0x0F) >= 0x0D:
            continue
        dr = nr - r
        dg = ng - g
        db = nb - b
        dist = dr*dr*3 + dg*dg*6 + db*db
        if dist < bestDist:
            bestColor = i
            bestDist = dist
    return bestColor

def getlut(img):
    """Get the LUT from a PIL type P image."""
    if img.mode != 'P':
        raise ValueError('image does not have a palette')
    lut = img.resize((256, 1))
    lut.putdata(range(256))
    return list(lut.convert("RGB").getdata())
    
# Pillow offers no built-in function to convert an image to use
# the closest color from another palette.
# I ran a gradient test image through
# imorig.convert('P', dither=Image.NONE, palette=nespal*4)
# and it looked almost as if it were first being quantized to
# web safe colors before being quantized to the NES palette.
# I ran an image with already very-close-to-NES colors through
# imorig.quantize(palette=nespalimg)
# and the result had what appeared to be Floyd-Steinberg
# dithering artifacts.
# So instead, I'll take the first four colors from the image,
# find the closest NES color to each, and convert it pixel-by-
# pixel that way.
# On the other hand, there's a Stack Overflow question about this.

def round_to_nes_colors(imorig):
    """Find the closest color in nespal to each color in a PIL image.

Return a list of tuples, one for each encountered color:
(index, (r, g, b), nes_color_index)

For an indexed image (mode "P"), index is the actual index.
For a truecolor image (mode "RGB"), index is the number of colors
with smaller luma.
The (r, g, b) refer to the original image, not to the closest NES
color, to allow converting true-color images.

"""
    origcolors = imorig.getcolors()
    if origcolors is None:
        raise ValueError("too many colors")
    if imorig.mode == 'P':
        def sortkey(frq_i):
            frq, i = frq_i
            return i
        origcolors.sort(key=sortkey)
        lut = getlut(imorig)
        rgbcolors = [(frq, i, lut[i]) for (frq, i) in origcolors]
    else:
        def sortkey(frq_r_g_b):
            frq, (r, g, b) = frq_r_g_b
            return 3*r+6*g+b
        origcolors.sort(key=sortkey)
        rgbcolors = [(frq, i, rgb)
                     for (i, (frq, rgb))
                     in zip(range(len(origcolors)), origcolors)]
    nescolors = [(i, rgb, nespal_closest(rgb)) for (frq, i, rgb) in rgbcolors]
    return nescolors

def convert_to_nes_colors(imorig, color0=None):
    """Convert an image to use four NES colors.

imorig -- a PIL image
color0 -- index into nespal (0x00-0x3F) to use as color 0

Return (new image, list of 4 NES colors)
Raise IndexError if there are too many colors in the image.

"""
    from PIL import Image
    from itertools import chain

    closest = round_to_nes_colors(imorig)

    # First, reorder the palette.
    # outpal is a list from tile values (0-3) to NES palette values
    # ready to write to $3F00-$3F03
    # mappings is a dict from values to outpal indices
    outpal = [color0, None, None, None]
    mappings = {}
    for (i, rgb, n) in closest:
        try:
            c = outpal.index(n)  # prefer existing colors
        except ValueError:
            if i < len(outpal) and outpal[i] is None:
                c = i  # prefer original color index
            else:
                try:
                    c = outpal.index(None)
                except ValueError:
                    raise ValueError("too many colors: %s, %02x"
                                     % (", ".join("%02x" % s for s in outpal),
                                        n))
            outpal[c] = n
        px = rgb if imorig.mode == 'RGB' else i
        mappings[px] = i

    out = Image.new('P', imorig.size)
    imdata = [mappings[i] for i in imorig.getdata()]
    out.putdata(imdata)
    pal = [nespal[i or 0] for i in outpal] + [nespal[0x0F] * 252]
    pal = list(chain(*pal))
    out.putpalette(pal)
    return (out, outpal)

def test():
    from PIL import Image

    filename = '../tilesets/screenshots/LAN Master.png'
    imorig = Image.open(filename)
    (newimg, newpal) = convert_to_nes_colors(imorig, 0x0F)
    closest = [(i, nespal[n], n) for (i, n) in enumerate(newpal)]    
    print("\n".join("%d (#%02x%02x%02x): $%02x" % (i, r, g, b, n)
                    for (i, (r, g, b), n) in closest))
    newimg.show()

if __name__=='__main__':
    test()
